/*
**   RAGGMOPP -- Copies files, replaces existing files only if
**               date stamp on source file is > date stamp on existing file
**
**       Thomas A. Lundin
**       Graphics Unlimited Inc.
**       3000 Second Street North
**       Minneapolis, MN 55411
**       (612) 588-7571
**
**    12/30/87
**    Version 1.2 -- Turbo C
**    1/6/88   v 1.3 -- stupid buffer size error fixed
**
*/

#define TRUE 1
#define FALSE 0

#include <stdio.h>
#include <ctype.h>
#include <conio.h>
#include <fcntl.h>
#include <dos.h>
#include <dir.h>
#include <io.h>
#include <stdlib.h>
#include <sys\stat.h>

#define BUFSZ 32767

char output_file[64];                     /* output file name */
char input_file[64];                      /* input file name */
char fname_buffer[128];                   /* line buffer for gang files */
char *fname_ptr;                          /* ptr for above */

char *data;                               /* where the file data are stored */
char *dptr;                               /* pointer to that area */

int fh,fi,fo;                             /* file handles */
int ctrlc();                              /* where CTRL-BREAK goes */

FILE *fl;

char s_path[80];                 /* source path */
char s_drv[3];
char s_dir[66];
char s_file[9];
char s_ext[5];

char t_path[80];                 /* target path */
char t_drv[3];
char t_dir[66];
char t_file[9];
char t_ext[5];

char cur_dir[66];
char t_chdir[66];

struct dfree freespace;
struct ftime ftod, ftod2;
struct ffblk fcb;
struct stat fst, fst2;

int query = FALSE;
int replace = FALSE;
char name[13];

main(argc, argv)
int argc;
char *argv[];
{
   int i, j, k;

   ctrlbrk(ctrlc);

   if(argc < 3)
      {
      puts("RAGGMOPP  ver.1.3 (c) 1987 Thomas A. Lundin");
      puts("          Copies files, replacing existing ones only if");
      puts("          the existing file has a less recent date");
      puts("usage:\n    mopp  [d:][\\path\\][filespec]  d:[\\path]  [-q] [-r]");
      puts("-or-\n    mopp  @dirlist  d:[\\path]  [-q] [-r]");
      puts("    'dirlist' is a file created by redirecting a DIR command");
      puts("    -q  forces a query before replacing existing files");
      puts("    -r  replaces existing files regardless of date");
      exit(1);
   }

   for (i = 3; (argc-3) > 0; i++, argc--) {    /* process options */
      if (strnicmp(argv[i],"-R",2) == 0) {
         replace = TRUE;
         fprintf(stderr,"Replacements will be allowed regardless of date\r\n");
      }
      else if (strnicmp(argv[i],"-Q",2) == 0) {
         query = TRUE;
         fprintf(stderr,"All replacements must be verified\r\n");
      }
      else {
         fprintf(stderr,"\r\nInvalid option: %s\r\n",argv[i]);
         exit(1);
      }
   }

   getcwd(cur_dir,66);                             /* get current dir */
   fnsplit(argv[1],s_drv,s_dir,s_file,s_ext);      /* split the names */
   fnsplit(argv[2],t_drv,t_dir,t_file,t_ext);

   strcpy(s_path,argv[1]);                         /* source pathname */
   strcpy(t_chdir,argv[2]);                        /* target dirname */

   if ((data = malloc(BUFSZ)) == NULL) {        /* space for data */
      fprintf(stderr,"Unable to grab data space\r\n");
      exit(1);
   }

/*
                                The main caller
*/

   if (*s_file == '@') {               /* process from a DIRLIST */
      strcpy(name,s_file+1);           /* form a file.name */
      strcat(name,s_ext);
      if ((fl = fopen(name,"r")) == NULL) {
         fprintf(stderr,"Can't open up %s\r\n",name);
         errorexit(1);
      }
      while ((fname_ptr = fgets(fname_buffer,128,fl)) != NULL) {
         if ((isalnum(*fname_ptr) || ispunct(*fname_ptr))
         && *fname_ptr != '.') {
            make_fn(t_path,fname_ptr);
            mopp(t_path);
         }
      }
   }
   else {                            /* process from a FILESPEC */
      i = 0;                         /* iteration sentinel */
      while (TRUE) {
         if (i == 0) {
            if (findfirst(s_path,&fcb,0) == EOF) {
               fprintf(stderr,"No files matching %s\r\n",s_path);
               errorexit(1);
            }
         }
         else {
            if (findnext(&fcb) == EOF) {
               break;
            }
         }
         ++i;
         mopp(fcb.ff_name);
      }
   }
   chdir(cur_dir);
   exit(0);
}

/************************
   The WORKHORSE ROUTINE
*************************/

mopp(in_fn)
char *in_fn;
{
   int ok_to_write;
   int non_existing;
   int length;
   char c;

   if ((t_chdir[2] & t_chdir[3]) != NULL) {
      if (access(t_chdir,0) != 0) {      /* need to MKDIR on target? */
         if (mkdir(t_chdir) == EOF) {
            fprintf(stderr,"Unable to create subdir %s\r\n",t_chdir);
            errorexit(1);
         }
      }
   }
   strcpy(input_file,s_drv);
   strcat(input_file,s_dir);
   strcat(input_file,in_fn);

   if ((fi = open(input_file,O_RDONLY | O_BINARY)) == EOF) {
      fprintf(stderr,"Unable to open input %s\r\n",input_file);
      errorexit(1);
   }
   stat(input_file,&fst);              /* get file stats */
   getftime(fi,&ftod);                 /* read the tod stamp */

   strcpy(output_file,t_chdir);
   strcat(output_file,"\\");
   strcat(output_file,in_fn);

   ok_to_write = FALSE;
   non_existing = TRUE;
   if (access(output_file,0) == 0) {      /* output already exists */
      non_existing = FALSE;
      fo = open(output_file, O_RDONLY);   /* open */
      stat(output_file,&fst2);            /* get file stats */
      getftime(fo,&ftod2);                /* read file stamp */
      _close(fo);                         /* shut */
      if (ft_comp() > 0)                  /* cur. file newer than existing? */
         ok_to_write = TRUE;              /* yes */
   }
   else {
      ok_to_write = TRUE;                 /* if output doesn't exist */
   }

   if ((replace) || (ok_to_write)) {
      c = '?';
      if(query) {
         fprintf(stderr,"Should  %-40.40s %10ld  %02d/%02d/%02d  %02d:%02d:%02d\r\n", input_file,fst.st_size,ftod.ft_month,ftod.ft_day,ftod.ft_year+80,ftod.ft_hour,ftod.ft_min,ftod.ft_tsec*2);
         if (non_existing)
            fprintf(stderr,"create  %-40.40s %10ld  %02d/%02d/%02d  %02d:%02d:%02d\r\n",output_file,fst.st_size,ftod.ft_month,ftod.ft_day,ftod.ft_year+80,ftod.ft_hour,ftod.ft_min,ftod.ft_tsec*2);
         else
            fprintf(stderr,"replace %-40.40s %10ld  %02d/%02d/%02d  %02d:%02d:%02d\r\n",output_file,fst2.st_size,ftod2.ft_month,ftod2.ft_day,ftod2.ft_year+80,ftod2.ft_hour,ftod2.ft_min,ftod2.ft_tsec*2);
         fprintf(stderr,"...(Y/N) ?  ");
         do {
            fprintf(stderr,"%c",0x08);
            c = toupper(getch());
         } while (c != 'Y' && c != 'N');
         fprintf(stderr,"%c\r\n",c);
      }
      if (query == FALSE || c == 'Y') {
         if ((fo = open(output_file,O_WRONLY | O_BINARY | O_CREAT, S_IREAD | S_IWRITE)) == EOF) {
            fprintf(stderr,"Can't open output %s\r\n",output_file);
            errorexit(1);
         }
         fprintf(stdout,"copy    %-40.40s %10ld  %02d/%02d/%02d  %02d:%02d:%02d\n",   input_file,fst.st_size,ftod.ft_month,ftod.ft_day,ftod.ft_year+80,ftod.ft_hour,ftod.ft_min,ftod.ft_tsec*2);
         if (non_existing)
            fprintf(stdout,"  as    %-40.40s %10ld  %02d/%02d/%02d  %02d:%02d:%02d\n\n",  output_file,fst.st_size,ftod.ft_month,ftod.ft_day,ftod.ft_year+80,ftod.ft_hour,ftod.ft_min,ftod.ft_tsec*2);
         else
            fprintf(stdout,"over    %-40.40s %10ld  %02d/%02d/%02d  %02d:%02d:%02d\n\n",output_file,fst2.st_size,ftod2.ft_month,ftod2.ft_day,ftod2.ft_year+80,ftod2.ft_hour,ftod2.ft_min,ftod2.ft_tsec*2);

         while ((length = _read(fi,data,BUFSZ)) > NULL)
            _write(fo,data,length);

         _close(fo);
         fo = open(output_file, O_RDONLY);        /*  open */
         setftime(fo,&ftod);                      /*  update time */
         _close(fo);                              /*  shut */
      }
   }
   else {
      fprintf(stdout,"source  %-40.40s %10ld  %02d/%02d/%02d  %02d:%02d:%02d\n",   input_file,fst.st_size,ftod.ft_month,ftod.ft_day,ftod.ft_year+80,ftod.ft_hour,ftod.ft_min,ftod.ft_tsec*2);
      fprintf(stdout,"target  %-40.40s %10ld  %02d/%02d/%02d  %02d:%02d:%02d\n",output_file,fst2.st_size,ftod2.ft_month,ftod2.ft_day,ftod2.ft_year+80,ftod2.ft_hour,ftod2.ft_min,ftod2.ft_tsec*2);
      fprintf(stdout,"** FILE NOT COPIED. Existing file has equal or more recent date. **\n\n");
   }
   _close(fi);
   return;
}

/*
             Make a true file name from a line in the DIRLIST
*/
make_fn(name,line)
char *line, *name;
{
    char temp[20];

    *name = NULL;                      /* init name buffer */
    stctok(line,temp,20," ");          /* get the name */
    strcat(name,temp);
    strcat(name,".");
    stctok(line+9,temp,20," ");        /* get the extension */
    strcat(name,temp);
    return;
}

/*
            Compare two file stamps
            return < 0 if stamp1 < stamp2
                   = 0 if stamp1 = stamp2
                   > 0 if stamp1 > stamp2
*/

ft_comp()
{
   int i;

   i = ftod.ft_year - ftod2.ft_year;
   if (i != 0) return (i);
   i = ftod.ft_month - ftod2.ft_month;
   if (i != 0) return (i);
   i = ftod.ft_day - ftod2.ft_day;
   if (i != 0) return (i);
   i = ftod.ft_hour - ftod2.ft_hour;
   if (i != 0) return (i);
   i = ftod.ft_min - ftod2.ft_min;
   if (i != 0) return (i);
   i = ftod.ft_tsec - ftod2.ft_tsec;
   return (i);
}

/*
            Copy a string up to a terminator character or length
*/
stctok(from,to,length,term)
char *from,*to,*term;
int length;
{
   char c;

/* The following statement will generate a Turbo C WARNING; pay no attention. */
   while((c = *from++) && (c != *term) && length--)
      *to++ = c;
   *to = NULL;
   return;
}

ctrlc()
{
   chdir(cur_dir);
   fprintf(stderr,"\r\nCancelled\r\n");
   return(0);
}

errorexit(n)
int n;
{
   chdir(cur_dir);
   exit(n);
}
